/*
FILE: LLCSYNC Reduction functions
CONTENTS: Simple reduction functions
LAST REV: 21/8/98
*/

#include "llcsync.h"


#define PREFIXBYADDV(sum, v, size) \
{ \
        int lli, llx, llw; \
	for (lli = 1; lli < size; lli++) \
		v[lli] = v[lli] + v[lli-1]; \
	sum = v[lli-1]; \
	for (llw = 0, lli = 0; lli < bit; lli++) \
	{ \
		if (LLNAME & (1 << lli)) \
		{ \
			SENDINT(lli, sum); \
			RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
			RECEIVEINT(lli, llx); \
			SENDINT(lli, sum); \
		} \
		sum += llx; \
		if (LLNAME & (1 << lli)) \
		llw += llx; \
	} \
	for (lli = 0; lli < size; lli++) \
		v[lli] += llw; \
}

#define PREFIXBYMULTV(mult, v, size) \
{ \
	int lli, llx, llw; \
	for (lli = 1; lli < size; lli++) \
		v[lli] = v[lli] * v[lli-1]; \
	mult = v[lli-1]; \
	for (llw = 1, lli = 0; lli < bit; lli++) \
	{ \
		if (LLNAME & (1 << lli)) \
		{ \
			SENDINT(lli, mult); \
			RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
		RECEIVEINT(lli, llx); \
			SENDINT(lli, mult); \
		} \
		mult *= llx; \
		if (LLNAME & (1 << lli)) \
      llw *= llx; \
	} \
	for (lli = 0; lli < size; lli++) \
		v[lli] *= llw; \
}

#define PREFIXBYMAXV(max, v, size) \
{ \
	int lli, llx, llw; \
	for (lli = 1; lli < size; lli++) \
		if (v[lli] < v[lli-1]) \
			v[lli] = v[lli-1]; \
	max = v[size-1]; \
	for (llw = max, lli = 0; lli < bit; lli++) \
	{ \
		if (LLNAME & (1 << lli))  \
		{ \
			SENDINT(lli, max); \
			RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
			RECEIVEINT(lli, llx); \
			SENDINT(lli, max); \
		} \
		if (max < llx) \
			max = llx; \
		if (LLNAME & (1 << lli)) \
      if (llw < max) \
        llw = max; \
	} \
	for (lli = 0; lli < size; lli++) \
		if (v[lli] > llw) \
			v[lli] = llw; \
}

#define PREFIXBYMINV(min, v, size) \
{ \
	int lli, llx, llw; \
	for (lli = 1; lli < size; lli++) \
		if (v[lli] > v[lli-1]) \
			v[lli] = v[lli-1]; \
	min = v[size-1]; \
	for (llw = min, lli = 0; lli < bit; lli++) \
	{ \
		if (LLNAME & (1 << lli))  \
		{ \
			SENDINT(lli, min); \
			RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
			RECEIVEINT(lli, llx); \
			SENDINT(lli, min); \
		} \
		if (min > llx) \
			min = llx; \
		if (LLNAME & (1 << lli)) \
		if (llw > llx) \
		  llw = llx; \
	} \
	for (lli = 0; lli < size; lli++) \
		if (v[lli] > llw) \
			v[lli] = llw; \
}

#define PREFIXBYADD(an, sum) \
{ \
	int llx, lli; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli)) \
		{ \
			SENDINT(lli, sum); \
                        RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llx); \
			SENDINT(lli, sum); \
		} \
		if ((LLNAME & (1 << lli)) != 0) \
			an = an + llx; \
		sum = sum + llx; \
	} \
}

#define PREFIXBYMULT(an, mult) \
{ \
	int llx, lli; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli)) \
		{ \
			SENDINT(lli, mult); \
                        RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llx); \
			SENDINT(lli, mult); \
		} \
		if ((LLNAME & (1 << lli)) != 0) \
			an = an * llx; \
		mult = mult * llx; \
	} \
}

#define PREFIXBYMIN(an, min) \
{ \
	int llx, lli; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli)) \
		{ \
			SENDINT(lli, min); \
                        RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llx); \
			SENDINT(lli, min); \
		} \
		if ((LLNAME & (1 << lli)) != 0) \
			if (llx < an) \
				an = llx; \
		if (llx < min) \
			min = llx; \
	} \
}

#define PREFIXBYMAX(an, max) \
{ \
	int llx, lli; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli)) \
		{ \
			SENDINT(lli, max); \
                        RECEIVEINT(lli, llx); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llx); \
			SENDINT(lli, max); \
		} \
		if ((LLNAME & (1 << lli)) != 0) \
			if (llx > an) \
				an = llx; \
		if (llx > max) \
			max = llx; \
	} \
}

#define REDUCEBYMIN(y,x) \
{ \
  int llaux, lli; \
  y = x; \
  for (lli = 0; lli < bit; lli++) \
  { \
		if ((LLNAME) & (1 << lli)) \
    { \
      SENDINT(lli, y); \
                        RECEIVEINT(lli, llaux); \
		} \
    else \
    { \
      RECEIVEINT(lli, llaux); \
      SENDINT(lli, y); \
    } \
    if (y > llaux) y = llaux; \
  } \
}

#define REDUCEBYMAX(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
								RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		if (y < llaux) y = llaux; \
	} \
}

#define REDUCEBYADD(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
                        RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		y += llaux; \
	} \
}

#define REDUCEBYMULT(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
                        RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		y *= llaux; \
	} \
}

#define REDUCEBYBAND(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
                        RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		y = y & llaux; \
	} \
}

#define REDUCEBYBOR(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
                        RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		y = y | llaux; \
	} \
}

#define REDUCEBYBXOR(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
                        RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		y = y ^ llaux; \
	} \
}

#define REDUCEBYLAND(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
                        RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
								RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		y = y && llaux; \
	} \
}

#define REDUCEBYLOR(y,x) \
{ \
	int llaux, lli; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			SENDINT(lli, y); \
								RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, y); \
		} \
		y = y || llaux; \
	} \
}

#define REDUCEBYMINV(min, v, size) \
{ \
	int lli, llaux; \
	min = v[0];  \
	for (lli = 1; lli < size; lli++) \
		if (v[lli] < min) \
			min = v[lli]; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli))  \
		{ \
			SENDINT(lli, min); \
			RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
			RECEIVEINT(lli, llaux); \
			SENDINT(lli, min); \
		} \
		if (llaux < min) min = llaux;\
	} \
}

#define REDUCEBYMAXV(max, v, size) \
{ \
	int lli, llaux; \
	max = v[0]; \
	for (lli = 1; lli < size; lli++) \
		if (max < v[lli]) \
			max = v[lli]; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli))  \
		{ \
			SENDINT(lli, max); \
			RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
			RECEIVEINT(lli, llaux); \
			SENDINT(lli, max); \
		} \
		if (max < llaux) max = llaux; \
	} \
}

#define REDUCEBYADDV(sum, v, size) \
{ \
	int lli, llaux; \
	sum = v[0]; \
	for (lli = 1; lli < size; lli++) \
		sum += v[lli]; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli))  \
		{ \
			SENDINT(lli, sum); \
			RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
			RECEIVEINT(lli, llaux); \
			SENDINT(lli, sum); \
		} \
		sum += llaux; \
	} \
}

#define REDUCEBYMULTV(mult, v, size) \
{ \
	int lli, llaux; \
	mult = v[0]; \
	for (lli = 1; lli < size; lli++) \
		mult *= v[lli]; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME & (1 << lli)) == (1 << lli))  \
		{ \
			SENDINT(lli, mult); \
			RECEIVEINT(lli, llaux); \
		} \
		else \
		{ \
			RECEIVEINT(lli, llaux); \
			SENDINT(lli, mult); \
		} \
		mult *= llaux; \
	} \
}


