/*
FILE: Complex reduction, broadcasting and routing macros.
LAST REV: 21/8/98
LAST CHANGES: (EWPBROADR): llx made dynamic array
*/

#ifndef LLCROUTEINCLUDED
#define LLCROUTEINCLUDED

#include "llcsync.h"

/* - Type of data - */
#ifndef TYPE
#   define TYPE int
#endif

/* - Operation to do in reduce macros - */
#ifndef OPERATION
#   define OPERATION(y, x) y = x + y
#endif

/* - Size of data (for supporting data arrays) - */
#ifndef SIZEOF
#   define SIZEOF(x) sizeof(x)
#endif


#define REDUCEDYN(y, size, x) \
{ \
  int lli, llaux; \
  TYPE *llaux_pointer; \
  size = SIZEOF(x); \
  memcpy(y,x,size);  \
	if ((llaux_pointer = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
    { \
      llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
      SEND(lli, y, llaux); \
      RECEIVEINT(lli, llaux); \
      RECEIVE(lli, llaux_pointer, llaux); \
    } \
    else \
    { \
                        RECEIVEINT(lli, llaux); \
			RECEIVE(lli, llaux_pointer, llaux); \
        llaux = SIZEOF(y); \
                        SENDINT(lli, llaux); \
                        SEND(lli, y, llaux); \
		} \
		OPERATION(y,llaux_pointer); \
                size = SIZEOF(y); \
	} \
	free (llaux_pointer); \
}

#define SREDUCEDYN(y, size, root, x) \
{ \
	int lli, llaux; \
	TYPE *llaux_pointer, *llresult; \
	if ((llaux_pointer = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	if ((llresult = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
        size = 0; \
	memcpy(llresult,x,SIZEOF(x));  \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
      if (!(root & (1 << lli))) \
      { \
        llaux = SIZEOF(llresult); \
                        SENDINT(lli, llaux); \
                        SEND(lli, llresult, llaux); \
      } \
      else \
      { \
        RECEIVEINT(lli, llaux); \
        RECEIVE(lli, llaux_pointer, llaux); \
    		OPERATION(llresult,llaux_pointer); \
      } \
    } \
    else \
    { \
      if (root & (1 << lli)) \
      { \
        llaux = SIZEOF(llresult); \
                        SENDINT(lli, llaux); \
                        SEND(lli, llresult, llaux); \
      } \
      else \
      { \
                       RECEIVEINT(lli, llaux); \
	  		RECEIVE(lli, llaux_pointer, llaux); \
    		OPERATION(llresult,llaux_pointer); \
      } \
		} \
	} \
	if (NAME == root) \
        { \
                size = SIZEOF(llresult); \
		memcpy(y,llresult,size);  \
         } \
	free (llaux_pointer); \
	free (llresult); \
}

#define REDUCE(y,x) \
{ \
	int lli, llaux; \
  TYPE llaux_pointer; \
  y = x; \
  for (lli = 0; lli < bit; lli++) \
  { \
		if ((LLNAME) & (1 << lli)) \
    { \
        llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
                        SEND(lli, &y, llaux); \
                        RECEIVEINT(lli, llaux); \
			RECEIVE(lli, &llaux_pointer, llaux); \
    } \
    else \
    { \
      RECEIVEINT(lli, llaux); \
      RECEIVE(lli, &llaux_pointer, llaux); \
      llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
      SEND(lli, &y, llaux); \
		} \
		OPERATION(y,llaux_pointer); \
	} \
}

#define SREDUCE(y, root, x) \
{ \
	int lli, llaux; \
	TYPE llaux_pointer, llresult; \
	llresult = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
      if (!(root & (1 << lli))) \
      { \
        llaux = SIZEOF(llresult); \
        SENDINT(lli, llaux); \
                          SEND(lli, &llresult, llaux); \
      } \
      else \
      { \
                        RECEIVEINT(lli, llaux); \
	  		RECEIVE(lli, &llaux_pointer, llaux); \
     		OPERATION(llresult,llaux_pointer); \
      } \
		} \
		else \
		{ \
      if (root & (1 << lli)) \
      { \
        llaux = SIZEOF(llresult); \
        SENDINT(lli, llaux); \
                          SEND(lli, &llresult, llaux); \
      } \
      else \
      { \
                        RECEIVEINT(lli, llaux); \
	  		RECEIVE(lli, &llaux_pointer, llaux); \
     		OPERATION(llresult,llaux_pointer); \
      } \
		} \
	} \
	if (NAME == root) \
		y = llresult; \
}

#define REDUCEBYNCODYN(y,size,x) \
{ \
	int lli, llaux; \
	TYPE *llaux_pointer; \
        size = SIZEOF(x); \
	memcpy(y,x,size); \
	if ((llaux_pointer = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	for (lli = 0; lli < bit; lli++) \
  { \
		if ((LLNAME) & (1 << lli)) \
		{ \
        llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
                        SEND(lli, y, llaux); \
      RECEIVEINT(lli, llaux); \
      RECEIVE(lli, llaux_pointer, llaux); \
			OPERATION(llaux_pointer, y); \
                        size = SIZEOF(llaux_pointer); \
			memcpy(y, llaux_pointer, size); \
		} \
    else \
    { \
      RECEIVEINT(lli, llaux); \
      RECEIVE(lli, llaux_pointer, llaux); \
      llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
      SEND(lli, y, llaux); \
      OPERATION(y, llaux_pointer); \
      size = SIZEOF(y); \
    } \
  } \
  free (llaux_pointer); \
}

#define SREDUCEBYNCODYN(y,size,root,x) \
{ \
	int lli, llaux; \
	TYPE *llaux_pointer, *llresult; \
	if ((llaux_pointer = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	if ((llresult = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
        size = 0; \
	memcpy(llresult,x,SIZEOF(x)); \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
      if (!(root & (1 << lli))) \
      { \
        llaux = SIZEOF(llresult); \
        SENDINT(lli, llaux); \
                          SEND(lli, llresult, llaux); \
      } \
      else \
      { \
                        RECEIVEINT(lli, llaux); \
	  		RECEIVE(lli, llaux_pointer, llaux); \
     		OPERATION(llaux_pointer,llresult); \
  			memcpy(llresult, llaux_pointer, SIZEOF(llaux_pointer)); \
      } \
		} \
		else \
		{ \
      if (root & (1 << lli)) \
      { \
        llaux = SIZEOF(llresult); \
        SENDINT(lli, llaux); \
                          SEND(lli, llresult, llaux); \
      } \
      else \
      { \
                        RECEIVEINT(lli, llaux); \
	  		RECEIVE(lli, llaux_pointer, llaux); \
     		OPERATION(llresult,llaux_pointer); \
      } \
		} \
	} \
	if (NAME == root) \
         { \
             size = SIZEOF(llresult); \
		memcpy(y,llresult,size); \
        } \
	free (llaux_pointer); \
	free (llresult); \
}

#define REDUCEBYNCO(y,x) \
{ \
	int lli, llaux; \
	TYPE llaux_pointer; \
	y = x; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
    { \
      llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
      SEND(lli, &y, llaux); \
      RECEIVEINT(lli, llaux); \
      RECEIVE(lli, &llaux_pointer, llaux); \
      OPERATION(llaux_pointer, y); \
      y = llaux_pointer; \
    } \
		else \
		{ \
      RECEIVEINT(lli, llaux); \
      RECEIVE(lli, &llaux_pointer, llaux); \
      llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
      SEND(lli, &y, llaux); \
      OPERATION(y, llaux_pointer); \
		} \
	} \
}

#define SREDUCEBYNCO(y,root,x) \
{ \
	int lli, llaux; \
	TYPE llaux_pointer, llresult; \
	llresult = x; \
  for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
    { \
      if (!(root & (1 << lli))) \
      { \
        llaux = SIZEOF(llresult); \
        SENDINT(lli, llaux); \
                          SEND(lli, &llresult, llaux); \
      } \
      else \
      { \
        RECEIVEINT(lli, llaux); \
        RECEIVE(lli, &llaux_pointer, llaux); \
     		OPERATION(llaux_pointer,llresult); \
        llresult = llaux_pointer; \
      } \
    } \
		else \
		{ \
      if (root & (1 << lli)) \
      { \
        llaux = SIZEOF(llresult); \
        SENDINT(lli, llaux); \
                          SEND(lli, &llresult, llaux); \
      } \
      else \
      { \
          RECEIVEINT(lli, llaux); \
	  		RECEIVE(lli, &llaux_pointer, llaux); \
     		OPERATION(llresult,llaux_pointer); \
      } \
		} \
	} \
	if (NAME == root) \
		y = llresult; \
}

#define SHIFTLEFT(message_receive, size, message_send) \
{ \
        int lli, llsize_receive, llaux; \
	BOOLEAN llrun; \
	TYPE *llaux_message; \
	if ((llaux_message = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	llrun = TRUE; \
	lli = 0; \
        size = SIZEOF (message_send); \
	memcpy(message_receive, message_send, size); \
	while (llrun) \
	{ \
		if ((NAME & (1 << lli)) == 0) \
		{ \
                        llaux = SIZEOF(message_receive); \
                        SENDINT (lli, llaux); \
                        SEND (lli, message_receive, llaux); \
                        RECEIVEINT (lli, llsize_receive); \
			RECEIVE (lli, message_receive, llsize_receive); \
                        size = llsize_receive; \
		} \
		else \
		{ \
                        RECEIVEINT (lli, llsize_receive); \
			RECEIVE (lli, llaux_message, llsize_receive); \
                        llaux = SIZEOF(message_receive); \
                        SENDINT (lli, llaux); \
                        SEND (lli, message_receive, llaux); \
                        size = SIZEOF(llaux_message); \
                        memcpy(message_receive, llaux_message, size); \
		} \
		if (((NAME & (1 << lli)) == 0) || (lli == (DIM - 1))) \
			llrun = FALSE; \
		else \
			lli = lli + 1; \
	} \
	free (llaux_message); \
}

#define SHIFTRIGHT(message_receive, size, message_send) \
{ \
        int lli, llsize_receive, llaux; \
	BOOLEAN llrun; \
	TYPE *llaux_message; \
	if ((llaux_message = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	llrun = TRUE; \
	lli = 0; \
        size = SIZEOF (message_send); \
	memcpy(message_receive, message_send, size); \
	while (llrun) \
	{ \
		if ((NAME & (1 << lli)) == 0) \
		{ \
                        llaux = SIZEOF(message_receive); \
                        SENDINT (lli, llaux); \
                        SEND (lli, message_receive, llaux); \
                        RECEIVEINT (lli, llsize_receive); \
			RECEIVE (lli, message_receive, llsize_receive); \
                        size = llsize_receive; \
		} \
		else \
		{ \
                        RECEIVEINT (lli, llsize_receive); \
			RECEIVE (lli, llaux_message, llsize_receive); \
                        llaux = SIZEOF(message_receive); \
                        SENDINT (lli, llaux); \
                        SEND (lli, message_receive, llaux); \
                        size = SIZEOF (llaux_message); \
			memcpy(message_receive, llaux_message, size); \
		} \
		if (((NAME & (1 << lli)) == (1 << lli)) || (lli == (DIM - 1))) \
			llrun = FALSE; \
		else \
			lli = lli + 1; \
	} \
	free (llaux_message); \
}

#define llpartner (LLNAME ^ (1 << lli))
#define HIGHPRIORITY(y, x) if (LLNAME < llpartner) strcpy(y, x)
#define LOWPRIORITY(y, x) if (LLNAME > llpartner) strcpy(y, x)

#define CWBROADR(y, size) \
{ \
  int lli, llaux, llempty; \
  TYPE *llaux_pointer; \
  if ((llaux_pointer = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
  llempty = 1; \
  size = 0; \
  for (lli = 0; ((lli < bit) && (llempty)); lli++) \
  { \
    if ((LLNAME) & (1 << lli)) \
    { \
      RECEIVEINT(lli, llaux); \
      if (llaux != 0) \
      { \
        size = llaux; \
        llempty = 0; \
        RECEIVE(lli, y, llaux); \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
    } \
    else \
    { \
      llaux = 0; \
      SENDINT(lli, llaux); \
      RECEIVEINT(lli, llaux); \
      if (llaux != 0) \
      { \
        size = llaux; \
        llempty = 0; \
        RECEIVE(lli, y, llaux); \
      } \
    } \
  } \
  for (;lli < bit; lli++) \
  { \
    if ((LLNAME) & (1 << lli)) \
    { \
      RECEIVEINT(lli, llaux); \
      if (llaux != 0) \
      { \
        RECEIVE(lli, llaux_pointer, llaux); \
        OPERATION (y, llaux_pointer); \
      } \
      llaux = SIZEOF(y); \
      size = llaux; \
      SENDINT(lli, llaux); \
      SEND(lli, y, llaux); \
    } \
    else \
    { \
      llaux = SIZEOF(y); \
      SENDINT(lli, llaux); \
      SEND(lli, y, llaux); \
      RECEIVEINT(lli, llaux); \
      if (llaux != 0) \
      { \
        RECEIVE(lli, llaux_pointer, llaux); \
        OPERATION (y, llaux_pointer); \
        size = SIZEOF(y); \
      } \
    } \
  } \
  free (llaux_pointer); \
}

#define CWBROADS(y, size, x) \
{ \
  int lli, llaux; \
  TYPE *llaux_pointer; \
  if ((llaux_pointer = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
  size = SIZEOF(x); \
  memcpy (y, x, size); \
  for (lli = 0; lli < bit; lli++) \
  { \
    if ((LLNAME) & (1 << lli)) \
    { \
      RECEIVEINT(lli, llaux); \
      if (llaux != 0) \
      { \
        RECEIVE(lli, llaux_pointer, llaux); \
        OPERATION (y, llaux_pointer); \
      } \
      llaux = SIZEOF(y); \
      size = llaux; \
      SENDINT(lli, llaux); \
      SEND(lli, y, llaux); \
    } \
    else \
    { \
      llaux = SIZEOF(y); \
      size = llaux; \
      SENDINT(lli, llaux); \
      SEND(lli, y, llaux); \
      RECEIVEINT(lli, llaux); \
      if (llaux != 0) \
      { \
        RECEIVE(lli, llaux_pointer, llaux); \
        OPERATION (y, llaux_pointer); \
        size = SIZEOF(y);\
      } \
    } \
  } \
  free (llaux_pointer); \
}

#define EWBROADR(y, size) \
{ \
	int lli, llaux, llempty, llsize; \
	llempty = 1; \
        size = 0; \
	for (lli = 0; ((lli < bit) && (llempty)); lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
                        RECEIVEINT(lli, llsize); \
			if (llsize != 0) \
			{ \
				llempty = 0; \
				RECEIVE(lli, y, llsize); \
                                size = llsize; \
			} \
                        llaux = 0; \
                        SENDINT(lli, llaux); \
		} \
		else \
		{ \
                        llaux = 0; \
                        SENDINT(lli, llaux); \
                        RECEIVEINT(lli, llsize); \
			if (llsize != 0) \
			{ \
				llempty = 0; \
				RECEIVE(lli, y, llsize); \
                                size = llsize; \
			} \
		} \
	} \
	for (;lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
                        RECEIVEINT(lli, llaux); \
			SENDINT(lli, llsize); \
			SEND(lli, y, llsize); \
		} \
		else \
		{ \
			SENDINT(lli, llsize); \
			SEND(lli, y, llsize); \
                        RECEIVEINT(lli, llaux); \
		} \
	} \
}

#define EWBROADS(x) \
{ \
	int lli, llaux; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
                        RECEIVEINT(lli, llaux); \
                        llaux = SIZEOF(x); \
                        SENDINT(lli, llaux); \
                        SEND(lli, x, llaux); \
		} \
		else \
		{ \
                        llaux = SIZEOF(x); \
                        SENDINT(lli, llaux); \
                        SEND(lli, x, llaux); \
                        RECEIVEINT(lli, llaux); \
		} \
	} \
}

#define EWPBROADR(y, size) \
{ \
	int lli, llj, llinc, llsize, llaux, llempty; \
	TYPE **llx; \
        llx = malloc(sizeof(TYPE *) * LLNUMPROCESSORS);\
	llinc = LLNAME; \
        size = 0; \
	for (lli = 0; lli < LLNUMPROCESSORS; lli++) \
	{ \
			if ((llx[lli] = (char *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
			llx[lli][0] = '\0'; \
	} \
	llempty = 1; \
	for (lli = 0; ((lli < bit) && (llempty)); lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
                        RECEIVEINT(lli, llaux); \
			if (llaux != 0) \
			{ \
				RECEIVE(lli, llx[llinc], llaux); \
				llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
				llsize = 1 << (bit - 1 - lli); \
				llempty = 0; \
				for (llj = 0; llj < (llsize-1); llj++) \
				{ \
                                        RECEIVEINT(lli, llaux); \
					RECEIVE(lli, llx[llinc], llaux); \
					llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
				} \
				llinc = LLNAME ^ (1 << (lli + 1)); \
			} \
                        llaux = 0; \
                        SENDINT(lli, llaux); \
		} \
		else \
		{ \
                        llaux = 0; \
                        SENDINT(lli, llaux); \
                        RECEIVEINT(lli, llaux); \
			if (llaux != 0) \
			{ \
				RECEIVE(lli, llx[llinc], llaux); \
				llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
				llsize = 1 << (bit - 1 - lli); \
				llempty = 0; \
				for (llj = 0; llj < (llsize-1); llj++) \
				{ \
                                        RECEIVEINT(lli, llaux); \
					RECEIVE(lli, llx[llinc], llaux); \
					llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
				} \
				llinc = LLNAME ^ (1 << (lli + 1)); \
			} \
		} \
	} \
	for (;lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
                        RECEIVEINT(lli, llaux); \
			llsize = 1 << (bit - 1 - lli); \
			for (llj = 0; llj < llsize; llj++) \
			{ \
                                llaux = SIZEOF(llx[llinc]); \
                                SENDINT(lli, llaux); \
                                SEND(lli, llx[llinc], llaux); \
				llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
			} \
			llinc = LLNAME ^ (1 << (lli + 1)); \
		} \
		else \
		{ \
			llsize = 1 << (bit - 1 - lli); \
			for (llj = 0; llj < llsize; llj++) \
			{ \
                                llaux = SIZEOF(llx[llinc]); \
                                SENDINT(lli, llaux); \
                                SEND(lli, llx[llinc], llaux); \
				llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
			} \
			llinc = LLNAME ^ (1 << (lli + 1)); \
                        RECEIVEINT(lli, llaux); \
		} \
	} \
        size = SIZEOF(llx[NAME]); \
	memcpy (y, llx[NAME], size); \
	for (lli = 0; lli < LLNUMPROCESSORS; lli++) \
		free(llx[lli]); \
        free(llx); \
}

#define EWPBROADS(x) \
{ \
	int lli, llj, llinc, llsize, llaux; \
	llinc = LLNAME ^ 1; \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
                        RECEIVEINT(lli, llaux); \
			llsize = 1 << (bit - 1 - lli); \
			for (llj = 0; llj < llsize; llj++) \
			{ \
                                llaux = SIZEOF(x[llinc]); \
                                SENDINT(lli, llaux); \
                                SEND(lli, x[llinc], llaux); \
				llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
			} \
			llinc = LLNAME ^ (1 << (lli + 1)); \
		} \
		else \
		{ \
			llsize = 1 << (bit - 1 - lli); \
			for (llj = 0; llj < llsize; llj++) \
			{ \
                                llaux = SIZEOF(x[llinc]); \
                                SENDINT(lli, llaux); \
                                SEND(lli, x[llinc], llaux); \
				llinc = (llinc + (1 << (lli + 1))) % NUMPROCESSORS; \
			} \
			llinc = LLNAME ^ (1 << (lli + 1)); \
                        RECEIVEINT(lli, llaux); \
		} \
	} \
}

#define EWROUTER(message, size) \
{ \
  int lli, llj, llaux; \
  LLROUTE *llstructaux, **llstruct; \
  TYPE *llmessageaux, **llmessage; \
  size = 0; \
  llstruct = malloc(sizeof(LLROUTE *) * DIM); \
  llmessage = malloc(sizeof(TYPE *) * DIM); \
  if ((llmessageaux = (TYPE *) malloc (SIZEBUFFER*sizeof(TYPE))) == NULL) abort(); \
  if ((llstructaux = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
  for (lli = 0; lli < bit; lli++) \
  { \
    if ((llstruct[lli] = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
    llstruct[lli]->procdest = -1; \
    if ((llmessage[lli] = (TYPE *) malloc (SIZEBUFFER*sizeof(TYPE))) == NULL) abort(); \
  } \
  for (lli = 0; lli < bit; lli++) \
  { \
    if ((LLNAME) & (1 << lli)) \
    { \
      for (llj = 0; llj < bit; llj++) \
      { \
        if (llstruct[llj]->procdest != -1) \
        { \
          if (!((llstruct[llj]->procdest) & (1 << lli))) \
          { \
            llaux = sizeof (*llstruct[llj]); \
            SENDINT(lli, llaux); \
            SEND(lli, llstruct[llj], llaux); \
            SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
            llstruct[llj]->procdest = -1; \
          } \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llstructaux, llaux); \
        RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
        if (llstructaux->procdest == LLNAME) \
        { \
           memcpy (message, llmessageaux, SIZEOF (llmessageaux)); \
           size = SIZEOF (llmessageaux); \
        } \
        else \
        { \
          for (llj = 0; llj < bit; llj++) \
          { \
            if (llstruct[llj]->procdest == -1) \
            { \
              memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
              memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
            } \
          } \
        } \
        RECEIVEINT(lli, llaux); \
      } \
    } \
    else \
    { \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llstructaux, llaux); \
        RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
        if (llstructaux->procdest == LLNAME) \
        { \
          memcpy (message, llmessageaux, SIZEOF (llmessageaux)); \
          size = SIZEOF (llmessageaux); \
        } \
        else \
        { \
          for (llj = 0; llj < bit; llj++) \
          { \
            if (llstruct[llj]->procdest == -1) \
            { \
              memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
              memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
            } \
          } \
        } \
        RECEIVEINT(lli, llaux); \
      } \
      for (llj = 0; llj < bit; llj++) \
      { \
        if (llstruct[llj]->procdest != -1) \
        { \
          if ((llstruct[llj]->procdest) & (1 << lli)) \
          { \
            llaux = sizeof (*llstruct[llj]); \
            SENDINT(lli, llaux); \
            SEND(lli, llstruct[llj], llaux); \
            SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
            llstruct[llj]->procdest = -1; \
          } \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
    } \
  } \
  for (lli = 0; lli < bit; lli++) \
  { \
    free (llmessage[lli]); \
    free (llstruct[lli]); \
  } \
  free (llmessageaux); \
  free (llstructaux); \
  free(llstruct); \
  free(llmessage); \
} \

#define EWROUTES(message_receive, size, procsend, message_send) \
{ \
  int lli, llj, llaux; \
  LLROUTE *llstructaux, **llstruct; \
  TYPE *llmessageaux, **llmessage; \
  size = 0; \
  llstruct = malloc(sizeof(LLROUTE *) * DIM); \
  llmessage = malloc(sizeof(TYPE *) * DIM); \
  if ((llmessageaux = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
  if ((llstructaux = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
  for (lli = 0; lli < bit; lli++) \
  { \
    if ((llstruct[lli] = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
    llstruct[lli]->procdest = -1; \
    if ((llmessage[lli] = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
  } \
  llstruct[0]->procdest = procsend; \
  llstruct[0]->sizemessage = SIZEOF (message_send); \
  memcpy (llmessage[0], message_send, SIZEOF (message_send)); \
  for (lli = 0; lli < bit; lli++) \
  { \
    if ((LLNAME) & (1 << lli)) \
    { \
      for (llj = 0; llj < bit; llj++) \
      { \
        if (llstruct[llj]->procdest != -1) \
        { \
          if (!((llstruct[llj]->procdest) & (1 << lli))) \
          { \
            llaux = sizeof (*llstruct[llj]); \
            SENDINT(lli, llaux); \
            SEND(lli, llstruct[llj], llaux); \
	    SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
	    llstruct[llj]->procdest = -1; \
          } \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llstructaux, llaux); \
        RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
        if (llstructaux->procdest == LLNAME) \
        { \
          size = SIZEOF (llmessageaux); \
          memcpy (message_receive, llmessageaux, size); \
        } \
        else \
        { \
          for (llj = 0; llj < bit; llj++) \
          { \
            if (llstruct[llj]->procdest == -1) \
            { \
              memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
              memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
            } \
          } \
        } \
        RECEIVEINT(lli, llaux); \
      } \
    } \
    else \
    { \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llstructaux, llaux); \
        RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
        if (llstructaux->procdest == LLNAME) \
        { \
          size = SIZEOF (llmessageaux); \
          memcpy (message_receive, llmessageaux, size); \
        } \
        else \
        { \
          for (llj = 0; llj < bit; llj++) \
          { \
            if (llstruct[llj]->procdest == -1) \
            { \
              memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
              memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
            } \
          } \
        } \
        RECEIVEINT(lli, llaux); \
      } \
      for (llj = 0; llj < bit; llj++) \
      { \
        if (llstruct[llj]->procdest != -1) \
        { \
          if ((llstruct[llj]->procdest) & (1 << lli)) \
          { \
            llaux = sizeof (*llstruct[llj]); \
            SENDINT(lli, llaux); \
            SEND(lli, llstruct[llj], llaux); \
            SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
            llstruct[llj]->procdest = -1; \
          } \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
    } \
  } \
  for (lli = 0; lli < bit; lli++) \
  { \
    free (llmessage[lli]); \
    free (llstruct[lli]); \
  } \
  free (llmessageaux); \
  free (llstructaux); \
  free(llstruct); \
  free(llmessage); \
} \

#define CWROUTER(message, size) \
{ \
  int lli, llj, llaux, llempty; \
  LLROUTE *llstructaux, **llstruct; \
  TYPE *llmessageaux, **llmessage; \
  size = 0; \
  llstruct = malloc(sizeof(LLROUTE *) * DIM); \
  llmessage = malloc(sizeof(TYPE *) * DIM); \
  if ((llmessageaux = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
  if ((llstructaux = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
  for (lli = 0; lli < bit; lli++) \
  { \
    if ((llstruct[lli] = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
    llstruct[lli]->procdest = -1; \
    if ((llmessage[lli] = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
  } \
  for (llempty = 1, lli = 0; lli < bit; lli++) \
  { \
    if ((LLNAME) & (1 << lli)) \
    { \
      for (llj = 0; llj < bit; llj++) \
      { \
        if (llstruct[llj]->procdest != -1) \
        { \
          if (!((llstruct[llj]->procdest) & (1 << lli))) \
          { \
            llaux = sizeof (*llstruct[llj]); \
            SENDINT(lli, llaux); \
            SEND(lli, llstruct[llj], llaux); \
            SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
            llstruct[llj]->procdest = -1; \
          } \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llstructaux, llaux); \
        RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
        if (llstructaux->procdest == LLNAME) \
        { \
          if (llempty) \
          { \
            memcpy (message, llmessageaux, SIZEOF (llmessageaux)); \
            size = SIZEOF (llmessageaux); \
          } \
          else \
          { \
            OPERATION (message, llmessageaux); \
          } \
          llempty = 0; \
        } \
        else \
        { \
          for (llj = 0; llj < bit; llj++) \
          { \
            if (llstruct[llj]->procdest == -1) \
            { \
              memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
              memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
            } \
          } \
        } \
        RECEIVEINT(lli, llaux); \
      } \
    } \
    else \
    { \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llstructaux, llaux); \
        RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
        if (llstructaux->procdest == LLNAME) \
        { \
          if (llempty) \
          { \
            memcpy (message, llmessageaux, SIZEOF (llmessageaux)); \
            size = SIZEOF (llmessageaux); \
          } \
          else \
          { \
            OPERATION (message, llmessageaux); \
          } \
          llempty = 0; \
        } \
        else \
        { \
          for (llj = 0; llj < bit; llj++) \
          { \
            if (llstruct[llj]->procdest == -1) \
            { \
              memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
              memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
            } \
          } \
        } \
        RECEIVEINT(lli, llaux); \
      } \
      for (llj = 0; llj < bit; llj++) \
      { \
        if (llstruct[llj]->procdest != -1) \
        { \
          if ((llstruct[llj]->procdest) & (1 << lli)) \
          { \
            llaux = sizeof (*llstruct[llj]); \
            SENDINT(lli, llaux); \
            SEND(lli, llstruct[llj], llaux); \
            SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
            llstruct[llj]->procdest = -1; \
          } \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
    } \
  } \
  for (lli = 0; lli < bit; lli++) \
  { \
    free (llmessage[lli]); \
    free (llstruct[lli]); \
  } \
  free (llmessageaux); \
  free (llstructaux); \
  free(llstruct); \
  free(llmessage); \
} \

#define CWROUTES(message_receive, size, procsend, message_send) \
{ \
	int lli, llj, llaux, llempty; \
	LLROUTE *llstructaux, **llstruct; \
	TYPE *llmessageaux, **llmessage; \
	size = 0; \
  llstruct = malloc(sizeof(LLROUTE *) * DIM); \
  llmessage = malloc(sizeof(TYPE *) * DIM); \
	if ((llmessageaux = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	if ((llstructaux = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
	for (lli = 0; lli < bit; lli++) \
	{ \
		if ((llstruct[lli] = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
		llstruct[lli]->procdest = -1; \
		if ((llmessage[lli] = (TYPE *) malloc (SIZEBUFFER * sizeof(TYPE))) == NULL) abort(); \
	} \
	llstruct[0]->procdest = procsend; \
	llstruct[0]->sizemessage = SIZEOF (message_send); \
	memcpy (llmessage[0], message_send, SIZEOF (message_send)); \
	for (llempty = 1, lli = 0; lli < bit; lli++) \
	{ \
		if ((LLNAME) & (1 << lli)) \
		{ \
			for (llj = 0; llj < bit; llj++) \
			{ \
				if (llstruct[llj]->procdest != -1) \
				{ \
					if (!((llstruct[llj]->procdest) & (1 << lli))) \
					{ \
                                                llaux = sizeof (*llstruct[llj]); \
                                                SENDINT(lli, llaux); \
                                                SEND(lli, llstruct[llj], llaux); \
						SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
						llstruct[llj]->procdest = -1; \
					} \
				} \
			} \
                        llaux = 0; \
                        SENDINT(lli, llaux); \
                        RECEIVEINT(lli, llaux); \
                        while (llaux != 0) \
			{ \
				RECEIVE(lli, llstructaux, llaux); \
				RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
				if (llstructaux->procdest == LLNAME) \
				{ \
					if (llempty) \
					{ \
						memcpy (message_receive, llmessageaux, SIZEOF (llmessageaux)); \
						size = SIZEOF (llmessageaux); \
					} \
					else \
					{ \
						OPERATION (message_receive, llmessageaux); \
					} \
					llempty = 0; \
				} \
				else \
				{ \
					for (llj = 0; llj < bit; llj++) \
					{ \
						if (llstruct[llj]->procdest == -1) \
						{ \
							memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
							memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
						} \
					} \
				} \
                               RECEIVEINT(lli, llaux); \
			} \
		} \
		else \
		{ \
                        RECEIVEINT(lli, llaux); \
                        while (llaux != 0) \
			{ \
				RECEIVE(lli, llstructaux, llaux); \
				RECEIVE(lli, llmessageaux, llstructaux->sizemessage); \
				if (llstructaux->procdest == LLNAME) \
				{ \
					if (llempty) \
					{ \
						memcpy (message_receive, llmessageaux, SIZEOF (llmessageaux)); \
						size = SIZEOF (llmessageaux); \
					} \
					else \
					{ \
						OPERATION (message_receive, llmessageaux); \
					} \
					llempty = 0; \
				} \
				else \
				{ \
					for (llj = 0; llj < bit; llj++) \
					{ \
						if (llstruct[llj]->procdest == -1) \
						{ \
							memcpy (llstruct [llj], llstructaux, sizeof (*llstructaux)); \
							memcpy (llmessage [llj], llmessageaux, SIZEOF (llmessageaux)); \
						} \
					} \
				} \
                                RECEIVEINT(lli, llaux); \
			} \
			for (llj = 0; llj < bit; llj++) \
			{ \
				if (llstruct[llj]->procdest != -1) \
				{ \
					if ((llstruct[llj]->procdest) & (1 << lli)) \
					{ \
                                                llaux = sizeof (*llstruct[llj]); \
                                                SENDINT(lli, llaux); \
                                                SEND(lli, llstruct[llj], llaux); \
						SEND(lli, llmessage[llj], SIZEOF (llmessage[llj])); \
						llstruct[llj]->procdest = -1; \
					} \
				} \
			} \
                        llaux = 0; \
                        SENDINT(lli, llaux); \
		} \
	} \
	for (lli = 0; lli < bit; lli++) \
	{ \
		free (llmessage[lli]); \
		free (llstruct[lli]); \
	} \
	free (llmessageaux); \
	free (llstructaux); \
  free(llstruct); \
  free(llmessage); \
} \

#define COLLECT(y, size, x) \
{ \
  int lli, llj, llaux; \
  LLROUTE **llsizemessage, *llsizemessageaux; \
  llsizemessage = malloc(sizeof(LLROUTE *) * NUMPROCESSORS); \
  if ((llsizemessageaux = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
  for (lli = 0; lli < (1 << bit); lli++) \
  { \
    if ((llsizemessage[lli] = (LLROUTE *) malloc (sizeof (LLROUTE))) == NULL) abort(); \
    llsizemessage[lli]->sizemessage = 0; \
    llsizemessage[lli]->procdest = -1; \
  } \
  llsizemessage[0]->sizemessage = SIZEOF(x); \
  llsizemessage[0]->procdest = LLNAME; \
  size[LLNAME] = SIZEOF(x); \
  memcpy(y[LLNAME],x,size[LLNAME]); \
  for (lli = 0; lli < bit; lli++) \
  { \
    if ((LLNAME) & (1 << lli)) \
    { \
      for (llj = 0; llj < (1 << lli); llj++) \
      { \
        if (llsizemessage[llj]->sizemessage != 0) \
        { \
          llaux = sizeof (*llsizemessage[llj]); \
          SENDINT(lli, llaux); \
          SEND(lli, llsizemessage[llj], llaux); \
          SEND(lli, y[llsizemessage[llj]->procdest], llsizemessage[llj]->sizemessage); \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llsizemessageaux, llaux); \
        RECEIVE(lli, y[llsizemessageaux->procdest], llsizemessageaux->sizemessage); \
        size[llsizemessageaux->procdest] = llsizemessageaux->sizemessage; \
        for (llaux = 1, llj = 0; ((llj < (1 << bit)) && (llaux)); llj++) \
          if (llsizemessage[llj]->sizemessage == 0) \
          { \
            memcpy(llsizemessage[llj], llsizemessageaux, sizeof (*llsizemessageaux)); \
            llaux = 0; \
          } \
          RECEIVEINT(lli, llaux); \
      } \
    } \
    else \
    { \
      RECEIVEINT(lli, llaux); \
      while (llaux != 0) \
      { \
        RECEIVE(lli, llsizemessageaux, llaux); \
        RECEIVE(lli, y[llsizemessageaux->procdest], llsizemessageaux->sizemessage); \
        size[llsizemessageaux->procdest] = llsizemessageaux->sizemessage; \
        for (llaux = 1, llj = 0; ((llj < (1 << bit)) && (llaux)); llj++) \
          if (llsizemessage[llj]->sizemessage == 0) \
          { \
            memcpy(llsizemessage[llj], llsizemessageaux, sizeof (*llsizemessageaux)); \
            llaux = 0; \
          } \
          RECEIVEINT(lli, llaux); \
      } \
      for (llj = 0; llj < (1 << lli); llj++) \
      { \
        if (llsizemessage[llj]->sizemessage != 0) \
        { \
          llaux = sizeof (*llsizemessage[llj]); \
          SENDINT(lli, llaux); \
          SEND(lli, llsizemessage[llj], llaux); \
          SEND(lli, y[llsizemessage[llj]->procdest], llsizemessage[llj]->sizemessage); \
        } \
      } \
      llaux = 0; \
      SENDINT(lli, llaux); \
    } \
  } \
  for (lli = 0; lli < (1 << bit); lli++) \
    free(llsizemessage[lli]); \
  free(llsizemessageaux); \
  free(llsizemessage); \
}

#endif /* #ifdef LLCROUTEINCLUDED */


